#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<mathlink.h>
#include"interval.h" 
const int size=20;

void recurse(int order,interval *x,interval *v,interval *y,interval *w,int n)
{
  int i,j;

  for(i=2;i<=order;i++) {
    x[i]=((2-n)*x[i-1]-v[i-1])/i;
    v[i]=(2.*v[i-1]+y[i-1])/i;
    y[i]=((4-n)*y[i-1]+w[i-1])/i;
    w[i]=interval(0.);
    for (j=0;j<i;j++) w[i]+=x[j]*w[i-1-j];
    w[i]=w[i]/i;
  }
}

void error(double *a,long a_len,int order,int n,double t)
{
  interval x[size],v[size],y[size],w[size];
  double b[8];

  x[0]=interval(a[0],a[1]);
  v[0]=interval(a[2],a[3]);
  y[0]=interval(a[4],a[5]);
  w[0]=interval(a[6],a[7]);
  x[1]=(2-n)*x[0]-v[0]+4*(n-2);
  v[1]=2.*v[0]+y[0];
  y[1]=(4-n)*y[0]+w[0];
  w[1]=x[0]*w[0];

  recurse(order,x,v,y,w,n);

  b[0]=x[order].left_b();
  b[1]=x[order].right_b();
  b[2]=v[order].left_b();
  b[3]=v[order].right_b();
  b[4]=y[order].left_b();
  b[5]=y[order].right_b();
  b[6]=w[order].left_b();
  b[7]=w[order].right_b();

  MLPutRealList(stdlink,b,8);
  return;
}

void dPhi(double *a,long a_len,int order,int n,double t)
{
  interval d[16][size],x[size],v[size],y[size],w[size],xi[16][size],ret[16];
  double b[32],p;
  int i,j,k,l,m;

  x[0]=interval(a[0],a[1]);
  v[0]=interval(a[2],a[3]);
  y[0]=interval(a[4],a[5]);
  w[0]=interval(a[6],a[7]);
  x[1]=(2-n)*x[0]-v[0]+4*(n-2);
  v[1]=2.*v[0]+y[0];
  y[1]=(4-n)*y[0]+w[0];
  w[1]=x[0]*w[0];
  recurse(order,x,v,y,w,n);

  for (i=0; i<16; i++) xi[i][0]=interval(0.);
  for (i=0; i<4; i++) xi[i+(i*4)][0]=interval(1.);
  d[0][0]=interval(-3.);
  d[1][0]=interval(-1.);
  d[2][0]=interval(0.);
  d[3][0]=interval(0.);
  d[4][0]=interval(0.);
  d[5][0]=interval(2.);
  d[6][0]=interval(1.);
  d[7][0]=interval(0.);
  d[8][0]=interval(0.);
  d[9][0]=interval(0.);
  d[10][0]=interval(-1.);
  d[11][0]=interval(1.);
  d[12][0]=w[0];
  d[13][0]=interval(0.);
  d[14][0]=interval(0.);
  d[15][0]=x[0];

  for (i=1; i<=order; i++) {
    for(j=0; j<16;j++) {
      if(j==15) d[j][i]=x[i];
      else d[j][i]=interval(0.);
    }
  }

  for (i=1; i<=order; i++) {
    for(k=0;k<4;k++) {
      for(m=0;m<4;m++) {
	xi[m+k*4][i]=interval(0.);
	for(j=0; j<i;j++) {
	  for(l=0;l<4;l++) xi[m+k*4][i]+=d[l+k*4][j]*xi[m+l*4][i-1-j];
	}
	xi[m+k*4][i]=xi[m+k*4][i]/i;
      }
    }
  }
  
  for (i=0; i<16; i++) ret[i]=interval(0.);

  for (k=0; k<=order; k++) {
    p=power(t,k);
    for (i=0; i<16; i++) ret[i]+=xi[i][k]*p;
  }

  for (i=0; i<16; i++) {
    b[i*2]=ret[i].left_b();
    b[i*2+1]=ret[i].right_b();
  }

  MLPutRealList(stdlink,b,32);
  return; 
}

void Phi(double *a,long a_len,int order,int n,double t)
{
  interval aa[4],x[size],v[size],y[size],w[size];
  double b[8],p;
  int i;

  x[0]=interval(a[0],a[1]);
  v[0]=interval(a[2],a[3]);
  y[0]=interval(a[4],a[5]);
  w[0]=interval(a[6],a[7]);
  x[1]=(2-n)*x[0]-v[0]+4*(n-2);
  v[1]=2.*v[0]+y[0];
  y[1]=(4-n)*y[0]+w[0];
  w[1]=x[0]*w[0];
 
  recurse(order,x,v,y,w,n);
  aa[0]=aa[1]=aa[2]=aa[3]=interval(0.);
  for (i=0; i<=order; i++) {
    p=power(t,i);
    aa[0]+=x[i]*p;
    aa[1]+=v[i]*p;
    aa[2]+=y[i]*p;
    aa[3]+=w[i]*p;
  }

  for (i=0; i<4; i++) {
    b[i*2]=aa[i].left_b();
    b[i*2+1]=aa[i].right_b();
  }
  
 MLPutRealList(stdlink,b,8);
 return;
}


interval f(interval *x,int i,int n)
{
  switch(i) {
  case 0:
    return((2-n)*x[0]-x[1]+4*(n-2));
  case 1:
    return(2.*x[1]+x[2]);
  case 2:
    return((4-n)*x[2]+x[3]);
  case 3:
    return(x[0]*x[3]);
  }
}

int in(interval *a, interval *b)
{
  int i,out;
  out=1;
  for (i=0;i<4;i++) {
  if ((a[i].left_b()>=b[i].left_b())||
      (a[i].right_b()<=b[i].right_b())) out=0;
    }
  return(out);
}

void Encl(double *a,long a_len,int n,double t)
{
  interval fat,x[4],y[4],z[4],zeroh;
  double b[8];
  int i;

  zeroh=interval(0.,t);
  fat=interval(-.005,.005);
  for(i=0;i<4;i++) x[i]=interval(a[2*i],a[2*i+1]);
  for(i=0;i<4;i++) z[i]=x[i]+zeroh*f(x,i,n)+fat;
  for(i=0;i<4;i++) y[i]=x[i]+zeroh*f(z,i,n);
  while(in(z,y)==0) {
    for(i=0;i<4;i++) {
      if ((z[i].left_b()>=y[i].left_b())||
      (z[i].right_b()<=y[i].right_b())) z[i]+=fat;
    }    
    for(i=0;i<4;i++) y[i]=x[i]+zeroh*f(z,i,n);
  }

  for (i=0; i<4; i++) {
    b[i*2]=y[i].left_b();
    b[i*2+1]=y[i].right_b();
  }
  
 MLPutRealList(stdlink,b,8);
 return;
}

