#include "interval.h"


#ifdef UP_ROUNDING
#  include "intv-upr.cpp"
#else
#  include "intv-swr.cpp"
#endif


bound_type EXP_ERR = 0;
bound_type SIN_ERR = 0;

extern const interval  INV_E( 2.7182818284590452, 2.7182818284590454 );
extern const interval     pi( 3.1415926535897932, 3.1415926535897934 );
extern const interval    pi2( 6.2831853071795864, 6.2831853071795867 ); // 2*pi
extern const interval  piby2( 1.5707963267948966, 1.5707963267948968 ); // pi/2
extern const interval pi3by2( 4.7123889803846898, 4.7123889803846902 ); // 3*pi/2




int  intersection(interval x1, interval x2,interval &cp)
 // on output:  if common part of <x1> and <x2> is nonempty then
 //                cp =intersection of <x1> and <x2>
 //                 and 1 is returned
 //              otherwise 0 is returned and cp is untouched
{
  int nonempty;
  double l=x1.left_b();
  if(x2.left_b() > l) l=x2.left_b();
  double r=x1.right_b();
  if(x2.right_b() < r) r=x2.right_b();

  nonempty=(l<=r);
  if(nonempty)  cp=interval(l,r);

  return(nonempty);
}




interval Ihull(const interval & i1, const interval & i2)
// the function returns an interval containing i1 and i2
{
  interval retval;
  retval.left=(i1.left < i2.left)? i1.left : i2.left;
  retval.right=(i1.right>i2.right)? i1.right: i2.right;
  return(retval);

}

double positive_power (double value, unsigned exponent, int round_dir)
// a computation of value^{exponent} with an option control of the direction
//   of rounding (if round_dir!=0)
//  if round_dir==0, then the computation is performed using current
//   rounding settings
{
  double result = 1;

  #ifndef UP_ROUNDING
  if (round_dir > 0)
    round_up ();
  else if (round_dir < 0)
    round_down ();
  #endif

  while (exponent)
  {
    while (!(exponent & 1))
    {
      exponent >>= 1;
      value *= value;
      #ifdef UP_ROUNDING
      if (round_dir < 0)
        dec_double (&value);
      #endif
    }
    exponent --;
    if (result == 1)
      result = value;
    else
    {
      result *= value;
      #ifdef UP_ROUNDING
      if (round_dir < 0)
        dec_double (&result);
      #endif
    }
  }

  return result;
}//--------------------------------------------------------------------

double power(double value, int exponent)
//  value^{exponent}
{
   return pow(value,exponent);

}


double power(int value, int exponent)
//  value^{exponent}
{
   return pow(value,exponent);

}//--------------------------------------------------------------------

interval power(const interval& value, int exponent)
//   value^{exponent}
{
   interval out=1;  // when   exponent==0
   double l,r,t;
   int prev_exp=exponent;
   int sign=1;

   if (exponent < 0 ) {
     sign=0;
     prev_exp=-prev_exp;
   }


   l=value.left;
   r=value.right;

   /* if 'exponent' is even, then we need to correct the ends of interval -
     only absolute values matter*/
   if (prev_exp % 2 == 0){

     l=fabs(l);
     r=fabs(r);

     if (l > r) {
       t=r;
       r=l;
       l=t;
     }

     if (in(0.0,value)) l=0.0;
   }

    if (r > 0)
      out. right = positive_power (r, prev_exp, 1);
    else if (r < 0)
      out. right = -positive_power (-r, prev_exp, -1);
    else
      out. right = 0;

    if (l > 0)
      out. left = positive_power (l, prev_exp, -1);
    else if (l < 0)
      out. left = -positive_power (-l, prev_exp, 1);
    else
      out. left = 0;

   if(sign==0)  // when an exponent is negative
   {
     out = 1/out;
   }

   return out;

}//--------------------------------------------------------------------

interval power(const interval &a, const interval &b)
{
  if( a.left_b() < 0 ) {
    cout << "pow(A, B): interval A must be greater than zero\n";
    exit(1);
  }
  return exp(b * log(a));
}

//--------------------------------- an absolute value
interval abs(const interval& _inter)
{
   interval temp=_inter;
   double l=fabs(temp.left);
   double r=fabs(temp.right);
   temp.left=(l<=r)?l:r;
   temp.right=(l>r)?l:r;
   if (in(0.0,_inter)) temp.left=0.0;
   return temp;

}//--------------------------------------------------------------------

//------  max of absolute value for an interval--------
double fabs(const interval& _inter)
{
   double temp;
   double l=fabs(_inter.left);
   double r=fabs(_inter.right);
   temp=(l<=r)?r:l;
   return temp;

}//--------------------------------------------------------------------

void iSplit(interval *i,double *diam)
// On output:  i(input) \subset i(output) + [-diam + diam]
{
 double l,r,s;
 l=(*i).left;
 r=(*i).right;
 s=(l+r)/2;
 *i=interval(s);
 #ifndef UP_ROUNDING
   round_up();
 #endif
 *diam=r-s;
 r=s-l;
 if (*diam<r)
   *diam=r;
}


//-----------an estimation for the remainder term during computation of "exp"
bound_type compute_EXP_ERR (void)
{
  // we start with a computation of the factorial, with rounding down because we divide by it later
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  double fctrial = 1;  // a factorial of i
  for (int i = 2; i <= EXP_TAYLOR + 1; i ++)
  {
    fctrial *= (double) i;
    #ifdef UP_ROUNDING
    // in fact this is necessary  only after some treshold (i > i_0), but we dont control it
    // hence we do it for all
    dec_double (&fctrial);
    #endif
  }
  // now we divide with up rounding
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  return INV_E. right_b () / fctrial;
} /* compute_EXP_ERR */

//--------------- An estimation of the remainder term for  "sin"
// It should be possible to use here the same approach as for the 'log'
// function, because we have here a series with alternating signs, and the
// absolute value of each term decreases to 0. Maybe we should do it this way
// in the future. - P. Pilarczyk
//
bound_type compute_SIN_ERR()
{
  // we start with a computation of the factorial, with rounding down because
  // we divide by it later
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  double fctrial = 1;
  for (int i = 2; i <= 2 * SIN_TAYLOR + 2; i ++)
  {
    fctrial *= (double) i;
    #ifdef UP_ROUNDING
     // in fact this is necessary  only after some treshold (i > i_0), but we dont control it
    // hence we do it for all
    dec_double (&fctrial);
    #endif
  }

  // now we compute powers and divisoon with rounding up
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  // the number 1.6 > pi/2 - this overestimates a little the remainder
  return positive_power (1.6, 2 * SIN_TAYLOR + 2, 0) / fctrial;
} /* compute_SIN_ERR */


//--------------- Trigonometric functions for interval arguments
//-------------------------------------------------- sin, cos, tan, cot

interval scaled_sin_2 (double x)
// a rigorous computation of sin(x) for 0<= x \leq \pi
//  if x<0 , then we compute sin(x=0)
{
  if (SIN_ERR == 0)
    SIN_ERR = compute_SIN_ERR ();
  if (x <= 0)
    return interval (0);
  interval h = power (interval (x) - piby2, 2);
  interval res (1);

  for(int i=SIN_TAYLOR; i>0; i--)
    res = 1 - ( h/(2*i * (2*i-1)) ) * res;


  interval retval;
  intersection(interval (-1, 1),res + interval (0, SIN_ERR),retval);

  return(retval);

} /* scaled_sin_2 */

interval scaled_sin_1 (const interval &x)
// the function computes sin(x) for x \subset [0,\infty]
{
  if (x. left_b () <= piby2. right_b ()) { // left end is to the left of pi/2
    if (x. right_b () >= pi3by2. left_b ())
      return interval (-1, 1); // because [Pi/2 , 3Pi/2] \subset x
    if (x. right_b () >= pi. left_b ()) //  [pi/2,Pi] \subset x, but x < 3*pi/2
      return interval
        (-(scaled_sin_2 ((x. right_b () - pi). right_b ()). right_b ()), 1);

    // now we know that   \pi/2 \in x and x < pi
    if (x. right_b () >= piby2. left_b ())
      return interval (min (scaled_sin_2 (x. left_b ()). left_b (),
        scaled_sin_2 (x. right_b ()). left_b ()), 1);

    // sin is an increasing function on x
    return interval (scaled_sin_2 (x. left_b ()). left_b (),
      scaled_sin_2 (x. right_b ()). right_b ());
  }
  //  we know that x > pi/2


  if (x. left_b () <= pi. right_b ()){ // Left(x) < pi
    if (x. right_b () >= (pi2 + piby2). left_b ())
      return interval (-1, 1); // because [pi,pi/2+2*pi] \subset x

    if (x. right_b () >= pi2. left_b ()) //   [pi,2*pi] \subset x
      return interval (-1, max (scaled_sin_2 (x. left_b ()). right_b (),
        scaled_sin_2 ((x - pi2). right_b ()). right_b ()));

    // now we know that  x < 2*pi and Left(x) < pi
    // hence  sinus achieves maximum on Left(x)
    double rsin = max (scaled_sin_2 (x. left_b ()). right_b (), 0.0);
    // if 2*pi/2 then minumum of sin(x)=-1
    if (x. right_b () >= pi3by2. left_b ())
      return interval (-1, rsin);

    // otherwise minimum is achieved at Right(x)
    if (x. right_b () >= pi. left_b ()) // we have to shift
      //  an argument to use scaled_sin_2
      return interval (-(scaled_sin_2 ((x. right_b () - pi). right_b ()).
        right_b ()), rsin);
    else
      return interval (max (scaled_sin_2 (x. right_b ()). left_b (), 0.0),
        rsin);
  }
  // lewy koniec przedzialu jest teraz na pewno na prawo od pi
  // we know that \pi < x

  return -scaled_sin_1 (x - pi);
} /* scaled_sin_1 */

interval sin (const interval &x)
{
  if (diameter (x). right_b () >= pi2. left_b ())
    return interval (-1, 1);

  // we shift x by 2*pi*k as close to zero as possible, we want x to be positive
  // after this shift
  long k;
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  if (x. left_b () > 0)
    k = (long) (x. left_b () / pi2. left_b ());
  else
    k = - (long) ((-x. left_b ()) / pi2. right_b () + 1);

  interval y = x - k * pi2;

  if (y. left_b () < 0)
    while (y. left_b () < 0)
      y += pi2;
  else
    while (y. left_b () >= pi2. right_b ())
      y -= pi2;

  if (diameter (y). right_b () >= pi2. left_b ())
    return interval (-1, 1);

  return scaled_sin_1 (y);
} /* sin */

interval cos(const interval &x)
{
  return sin(piby2-x);
}

interval tan(const interval &x)
{
  return sin(x)/sin(piby2-x);
}

interval cot(const interval& x)
{
  return sin(piby2-x)/sin(x);
}

//------------------------ Interval hyperbolic functions
//----------------------------- sinh, cosh, tanh, coth
interval cosh(const interval &x)
{
  return ( (exp(x)+exp(-x))/2 );
}

interval sinh(const interval &x)
{
  return ( (exp(x)-exp(-x))/2 );
}

interval tanh(const interval &x)
{
  return ( (exp(x)-exp(-x))/(exp(x)+exp(-x)) );
}

interval coth(const interval &x)
{
  return ( (exp(x)+exp(-x))/(exp(x)-exp(-x)) );
}


//------------------ an exponent function
double scaled_exp_up (double x)
// returns an upper bound for exp(x) for x in [0,1]
{
  double res = 0;

  #ifndef UP_ROUNDING
  round_up ();
  #endif
  for (int i = EXP_TAYLOR; i > 0; i --)
    res = (res + 1.0) * (x / i);
  return res + 1 + EXP_ERR;
} /* scaled_exp_up */

double scaled_exp_down (double x)
// a lower bound for exp(x) , for x in [0,1]
{
  double res = 0;

  #ifndef UP_ROUNDING
  round_down ();
  #endif
  for (int i = EXP_TAYLOR; i > 0; i --)
  {
    // res = (res + 1) * (x / i)
    res += 1.0;
    #ifdef UP_ROUNDING
    dec_double (&res);
    #endif
    double temp = x / i;
    #ifdef UP_ROUNDING
    dec_double (&temp);
    #endif
    res *= temp;
  }
  res += 1.0;
  #ifdef UP_ROUNDING
  dec_double (&res);
  #endif
  return res;
} /* scaled_exp_down */

interval exp (const interval &x)
{
  if (EXP_ERR == 0)
    EXP_ERR = compute_EXP_ERR ();

  // we compute first upper bound - this depends on Right(x)

  // we decompose Right(x)= right_int + right_diff
  //   where right_int is an integer
  //         right_diff - a fractional part - in [0,1]
  double right_x = x. right_b ();
  int right_int = (int) right_x;
  while (right_int > right_x)
    right_int --;
  // a fractional part of Right(x)
  #ifndef UP_ROUNDING
  round_up();
  #endif
  double right_diff = right_x - right_int;
  double e;
  if (right_int >= 0)
    e = INV_E. right_b ();
  else
  {
    e = 1.0 / INV_E. left_b ();
    right_int = -right_int;
  }
  double right_res = positive_power (e, right_int, 1) *
    scaled_exp_up (right_diff);



  // now we compute lower bound - this depends on Left(x)
  //   we have to round verious things down
  double left_x = x. left_b ();
  int left_int = (int) left_x;
  while (left_int > left_x)
    left_int --;
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  double left_diff = left_x - left_int;
  #ifdef UP_ROUNDING
  dec_double (&left_diff);
  #endif
  if (left_int >= 0)
    e = INV_E. left_b ();
  else
  {
    e = 1.0 / INV_E. right_b ();
    #ifdef UP_ROUNDING
    dec_double (&e);
    #endif
    left_int = -left_int;
  }
  double left_res = positive_power (e, left_int, -1) *
    scaled_exp_down (left_diff);
  #ifdef UP_ROUNDING
  dec_double (&left_res);
  #endif

  return interval (left_res, right_res);
} /* exp */

//-------------------------------the natural logarithm function
double scaled_log_up (double x)
// computes upper bound for log(1+x) for x \in [0,1/2]
// the constant LOG_TAYLOR must be even
{
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  double result = 1.0;
  result /= (LOG_TAYLOR + 1);
  for (int i = LOG_TAYLOR; i > 0; i -= 2)
    result = (result * x + (-1.0) / i) * x + 1.0 / (i - 1);
  return result * x;
} /* scaled_log_up */

double scaled_log_down (double x)
// computes lowe bound for log(1+x) for x \in [0,1/2]
// the constant LOG_TAYLOR must be even
{
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  double result = 0;
  for (int i = LOG_TAYLOR; i > 0; i -= 2)
  #ifdef UP_ROUNDING
  {
    result *= x;
    dec_double (&result);
    result *= x;
    dec_double (&result);
    double temp = 1.0 / (i - 1);
    dec_double (&temp);
    result += temp;
    dec_double (&result);
    temp = (-x) / i;
    dec_double (&temp);
    result += temp;
    dec_double (&result);
  }
  result *= x;
  dec_double (&result);
  #else
    result = result * x * x + 1.0 / (i - 1) + (-x) / i;
  return result * x;
  #endif
} /* scaled_log_down */

interval log (const interval &x)
// computes a natural logoraithm of x , x>0
{
  if( x.right_b()<=0 ) {
     cerr << "log(): domain error.\n";
     exit(1);
  }
  static const interval one_and_a_half (3.0 / 2, 3.0 / 2);
  static const interval log_one_and_a_half( 0.4054651081081643,
    0.4054651081081645);

  // we take an advantage of monotonicty

  // an upper estimate -
  //  we find an integer 'k' such that Right(x)=(1.5)^k*d , where d is in [1,1.5]
  int k = (int) (log (x. right_b ()) / log (1.5));

  // we find a lower bound for  (1.5)^k
  double y;
  if (k >= 0)
    y = positive_power (3.0 / 2, k, -1);
  else
  {
    y = positive_power (3.0 / 2, -k, 1);
    #ifndef UP_ROUNDING
    round_down ();
    #endif
    y = 1.0 / y;
    #ifdef UP_ROUNDING
    dec_double (&y);
    #endif
  }

  // try to increase k if possible
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  double y_prim = y * (3.0 / 2);
  #ifdef UP_ROUNDING
  dec_double (&y_prim);
  #endif
  while (y_prim < x. right_b ())
  {
    k ++;
    y = y_prim;
    y_prim = y * (3.0 / 2);
    #ifdef UP_ROUNDING
    dec_double (&y_prim);
    #endif
  }

  // decrease k if too big
  while (y > x. right_b ())
  {
    k --;
    y /= (3.0 / 2);
    #ifdef UP_ROUNDING
    dec_double (&y);
    #endif
  }

  // computes upper bound for d
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  double d = x. right_b () / y;

  // computation of an upper bound for k*log(1.5) + log(d)
  double right_result = scaled_log_up (d - 1);
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  right_result += ((k > 0) ? log_one_and_a_half. right_b () :
    log_one_and_a_half. left_b ()) * k;


  // teraz to samo dla lewego konca przedzialu:
  // now we proceed analogously with the left end
  k = (int) (log (x. left_b ()) / log (1.5));

  // upper bound for (1.5)^k
  if (k >= 0)
    y = positive_power (3.0 / 2, k, 1);
  else
  {
    y = positive_power (3.0 / 2, -k, -1);
    #ifndef UP_ROUNDING
    round_up ();
    #endif
    y = 1.0 / y;
  }

  // increase k if possiblle
  #ifndef UP_ROUNDING
  round_up ();
  #endif
  y_prim = y * (3.0 / 2);
  while (y_prim < x. left_b ())
  {
    k ++;
    y = y_prim;
    y_prim = y * (3.0 / 2);
  }

  // decrease k if too large
  while (y > x. left_b ())
  {
    k --;
    y /= (3.0 / 2);
  }

  // computation of lower bound for d
  #ifndef UP_ROUNDING
  round_down ();
  #endif
  d = x. left_b () / y;
  #ifdef UP_ROUNDING
  dec_double (&d);
  #endif

  // computation of lower bound for   log (1.5) * k + log d
  double d_1 = d - 1;
  #ifdef UP_ROUNDING
  dec_double (&d_1);
  #endif
  double left_result = scaled_log_down (d_1);
  #ifdef UP_ROUNDING
  dec_double (&left_result);
  #else
  round_down ();
  #endif
  double product = ((k > 0) ? log_one_and_a_half. left_b () :
    log_one_and_a_half. right_b ()) * k;
  #ifdef UP_ROUNDING
  dec_double (&product);
  #endif
  left_result += product;
  #ifdef UP_ROUNDING
  dec_double (&left_result);
  #endif

  return interval (left_result, right_result);
} /* log */

//---------- reading from the stream
 istream &operator>>(istream &inp, interval &i){
   int ch;
   bound_type ll,rr;
   i.left=0.0;i.right=0.0;
   while('['!=(ch=inp.get()) && ch!=EOF);
   if(ch!=EOF){
      inp >> ll;
      if(inp.get()==','){
        inp >> rr;
        if(inp.get()==']'){
          if(ll <= rr){
            i.left=ll;i.right=rr;
          }
        }
      }
   }
   return inp;
 }


 // solves inclusion a+[0,t]*p\subset c for t
interval solve_affine_inclusion(interval &a, interval & p, interval& c){
   interval t;
   if ( !(a.subset(c)) ){
      #ifdef DEBUGGING
      errorExit("Cannot solve affine inclusion");
      #endif
      //     return -1;
   }else{
     if(p>=0) t=interval(((interval(c.right)-interval(a.right))/interval(p.right)).left);
     else
       if(p<=0) t=interval(((interval(c.left)-interval(a.left))/interval(p.left)).left);
       else{
         double t1=((interval(c.right)-interval(a.right))/interval(p.right)).left;
         double t2=((interval(c.left)-interval(a.left))/interval(p.left)).left;
         t=interval( ( t1<t2 ? t1 : t2 ) );
       }
   }
   return(t);
 }

 interval solve_affine_inclusion(const interval &a, const interval &p, const interval &c, int &dir){
   interval t;
   if ( !(a.subset(c)) ){
     #ifdef DEBUGGING
       errorExit("Cannot solve affine inclusion");
     #endif
   }else{
     if(p>=0){
       dir=1;
       t=interval(((interval(c.right)-interval(a.right))/interval(p.right)).left);
     }else{
       if(p<=0){
         dir=-1;
         t=interval(((interval(c.left)-interval(a.left))/interval(p.left)).left);
       }else{
         double t1=((interval(c.right)-interval(a.right))/interval(p.right)).left;
         double t2=((interval(c.left)-interval(a.left))/interval(p.left)).left;
         if(t1<t2){
           dir=1;
           t=interval(t1);
         }else{
           dir=-1;
           t=interval(t2);
         }
       }
     }
   }
   return(t);
 }


#ifdef KRAK
  void interval_error(char *c, double left, double right){
    errorExit("Interval error: %s\nleft=%s right=%s",
      c,double2hex(left).string(),double2hex(right).string());
  }
#else
  void interval_error(char *c, double left, double right){
     cerr << "Interval error: " << c << "left=" << left << " right=" << right;
  }
#endif
void interval_test(interval &iv, char *comment){
  if(iv.left_b()>iv.right_b()) interval_error(comment, iv.left_b(), iv.right_b());
}


#ifdef KRAK
frstring &operator>>(frstring &f, interval &i){
  istrstream inp(f.string());
  streampos pos=inp.tellg();

  inp >> i;
  int k=inp.tellg()-pos;
  f=f(k);
  return f;
}
#endif


