//------------
//    Function for the class 'Interval' without switching the rounding
//    We should initiate rounding up by uncommenting #UP_ROUNDING
//        in file "setting.H'
//------------


unsigned int _state_word=0;      //  a word describing the state of coprocessor

//------------------------------
interval& interval::operator=(const interval& _inter)
{   if(this == &_inter)    // assignment a=a
       return *this;
    left  = _inter.left;
    right = _inter.right;
    return *this;
}//--------------------------------------------------------------------

//------------------- operator +=
interval& interval::operator+=(const interval& _inter)
{   left  += _inter.left;
    right += _inter.right;
    dec_double(&left);
    #ifdef  DEBUGGING
    if(left>right)
      interval_error(" Operator += ",left,right);
    #endif
    return *this;
}//--------------------------------------------------------------------

//--------------------------------------
interval& interval::operator-=(const interval& _inter)
{
    left  -= _inter.right;
    right -= _inter.left;
    dec_double(&left);
    #ifdef  DEBUGGING
    if(left>right)
      interval_error(" Operator -= ",left,right);
    #endif
    return *this;
}//--------------------------------------------------------------------

//--------------------------------------
interval& interval::operator*=(const interval& _inter)
{

    if((right==0.0) && (left==0.0)){
      return *this;
    }

    if((_inter.right==0.0) && (_inter.left==0.0)){
      left=0;
      right=0;
      return *this;
    }

    if (right<=0 && _inter.right<=0)       // (--)*(--)
    {   bound_type temp=right*_inter.right;
        right=left*_inter.left;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (--)(--)",left,right);
        #endif
        return *this;
    }


    if( right<=0 && _inter.left<=0 && _inter.right>=0)   // (-,-)*(-,+)
    {   bound_type temp=left*_inter.right;
        right=left*_inter.left;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (--)(-+)",left,right);
        #endif
        return *this;
    }

    if  ( right<=0 && _inter.left>=0)                      // (-,-)*(+,+)
    {   left*=_inter.right;
        right*=_inter.left;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (--)(++)",left,right);
        #endif
        return *this;
    }



    if(left<=0 && right>=0 && _inter.right<=0)           // (-,+)*(-,-)
    {   bound_type temp=right*_inter.left;
        right=left*_inter.left;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (-+)(--)",left,right);
        #endif
        return *this;
    }


                                                        // (-,+)*(-,+)
    if(left<=0 && right>=0 && _inter.left<=0 && _inter.right>=0)
    {  bound_type l1=left*_inter.right, l2=right*_inter.left;
       bound_type r1=left*_inter.left, r2=right*_inter.right;
       if(l1>l2) left=l2;
       else left=l1;
       if(r1<r2) right=r2;
       else right=r1;
       dec_double(&left);
       #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (-+)(-+)",left,right);
        #endif
       return *this;
    }



    if(left<=0 && right>=0 && _inter.left>=0)            // (-,+)*(+,+)
    {   left*= _inter.right;
        right*= _inter.right;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (-+)(++)",left,right);
        #endif
        return *this;
    }



    if(left>=0 && _inter.right<=0)                       // (+,+)*(-,-)
    {   bound_type temp=right*_inter.left;
        right=left*_inter.right;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (++)(--)",left,right);
        #endif
        return *this;
    }

    if(left>=0 && _inter.left<=0 && _inter.right>=0)     // (+,+)*(-,+)
    {   left=right*_inter.left;
        right*=_inter.right;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator *= (++)(-+)",left,right);
        #endif
        return *this;
    }


  if(left>=0 && _inter.right >=0){ // (++)(++)
    left*=_inter.left;
    right*=_inter.right;
    dec_double(&left);
    #ifdef  DEBUGGING
    if(left>right)
      interval_error(" Operator *= (++)(++)",left,right);
    #endif
    return *this;
  }





  cerr << "Error *** Fatal interval error." << endl;
  exit(1);

  return *this;
}//--------------------------------------------------------------------

//--------------------------------------
interval& interval::operator/=(const interval& _inter)
{
  if(_inter.left<=0 && _inter.right>=0){
    cerr << "Interval error : Possible divide by zero.";
    exit(1);
  }


  if( right<=0 && _inter.right<0)                      // (--)/(--)
  {
    bound_type rr=left/_inter.right;
    bound_type ll=right/_inter.left;
    left=ll;
    right=rr;
    dec_double(&left);
    #ifdef  DEBUGGING
    if(left>right)
      interval_error(" Operator /= (--)(--)",left,right);
    #endif
    return *this;
  }


   if( right<=0 && _inter.left>0)                       // (-,-)/(+,+)
    {   left/=_inter.left;
        right/=_inter.right;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator /= (--)(++)",left,right);
        #endif
        return *this;
    }

   if(left<=0 && right>=0 && _inter.right<0)            // (-,+)/(-,-)
    {   bound_type temp=right/_inter.right;
        right=left/_inter.right;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator /= (-+)(--)",left,right);
        #endif
        return *this;
    }




    if(left<=0 && right>=0 && _inter.left>0)             // (-,+)/(+,+)
    {   left /=_inter.left;
        right/=_inter.left;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator /= (-+)(++)",left,right);
        #endif
        return *this;
    }


    if(left>=0 && _inter.right<0)                        // (+,+)/(-,-)
    {   bound_type temp=right/_inter.right;
        right=left/_inter.left;
        left=temp;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator /= (++)(--)",left,right);
        #endif
        return *this;
    }


    if(left>=0 && _inter.left>0)                         // (+,+)/(+,+)
    {   left/=_inter.right;
        right/=_inter.left;
        dec_double(&left);
        #ifdef  DEBUGGING
        if(left>right)
          interval_error(" Operator /= (++)(++)",left,right);
        #endif
        return *this;
    }


    cerr << "Error *** Fatal interval error.   /=" << endl;
    exit(1);




  return *this;
}//--------------------------------------------------------------------

interval operator +(const interval& _inter1, const interval& _inter2)
{   bound_type ll,rr;

    ll = _inter1.left+_inter2.left;
    dec_double(&ll);
    rr = _inter1.right+_inter2.right;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator +(a) ",ll,rr);
        #endif
    return interval(ll,rr);
}

interval operator+(const interval& _inter, const bound_type _val)
{   bound_type ll,rr;

    ll = _inter.left+_val;
    dec_double(&ll);
    rr = _inter.right+_val;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator +(b)",ll,rr);
        #endif
   return interval(ll,rr);
}

interval operator+(const bound_type _val,const interval& _inter)
{   bound_type ll,rr;

    ll = _inter.left+_val;
    dec_double(&ll);
    rr = _inter.right+_val;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator +(c)",ll,rr);
        #endif
    return interval(ll, rr);
}//--------------------------------------------------------------------

//Operator odejmowania -
interval operator -(const interval& _inter1, const interval& _inter2)
{  bound_type ll,rr;

   ll = _inter1.left-_inter2.right;
   dec_double(&ll);
   rr = _inter1.right-_inter2.left;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator -(a)",ll,rr);
        #endif
   return interval(ll,rr);
}

interval operator-(const interval& _inter, const bound_type _val)
{  bound_type ll,rr;

   ll = _inter.left-_val;
   dec_double(&ll);
   rr = _inter.right-_val;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator -(b)",ll,rr);
        #endif
   return interval(ll,rr);
}

interval operator-(const bound_type _val,const interval& _inter)
{  bound_type ll,rr;

   ll = _val-_inter.right;
   dec_double(&ll);
   rr = _val-_inter.left;
        #ifdef  DEBUGGING
        if(ll>rr)
          interval_error(" Operator -(c)",ll,rr);
        #endif
   return interval(ll,rr);
}//--------------------------------------------------------------------



//-------------------------------------------------
interval  operator *(const interval& _inter1, const interval& _inter2){
    bound_type ll,rr,t;

    #ifdef DEBUGGING
    if(_inter1.left>_inter1.right)
      interval_error(" Operator * (1)",_inter1.left,_inter1.right);
    #endif
     #ifdef DEBUGGING
    if(_inter2.left>_inter2.right)
      interval_error(" Operator * (2)",_inter2.left,_inter2.right);
    #endif


    if((_inter1.left==0) && (_inter1.right==0)){
      return(interval(0.0,0.0));
    }

    if((_inter2.left==0) && (_inter2.right==0)){
      return(interval(0.0,0.0));
    }

    if (((_inter1.left >= 0) && (_inter2.left >=0)) || ((_inter1.right<=0)
           && (_inter2.right <= 0)) ){
               // (++)(++) ; (--)(--)
      ll=_inter1.left*_inter2.left;
      rr=_inter1.right*_inter2.right;
      if (ll >= rr){
        t=rr;
        rr=ll;
        ll=t;
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator * (++)(++) or (--)(--)",ll,rr);
        #endif
      }
    }else
    if ((_inter1.right <=0) && (_inter2.left <= 0)){ // (--)(-+)
                     // _inter2.right > 0 - from previous condition
      ll=_inter1.left*_inter2.right;
      rr=_inter1.left*_inter2.left;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (b)",ll,rr);
      #endif
    } else
    if ((_inter1.right <=0) && (_inter2.left >=0)){ // (--)(++)
      ll=_inter1.left*_inter2.right;
      rr=_inter1.right*_inter2.left;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (c)",ll,rr);
      #endif
    } else    // _inter1.right > 0 - from previous conditions
    if ((_inter1.left <=0) && (_inter2.right <= 0)) { // (-+)(--)
      ll=_inter1.right*_inter2.left;
      rr=_inter1.left*_inter2.left;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (d)",ll,rr);
      #endif
    } else
    if ((_inter1.left <=0) && (_inter2.left <=0) ){ // (-+)(-+)
      ll=_inter1.left*_inter2.right;
      t=_inter2.left*_inter1.right;
      if (t<ll) ll=t;
      rr=_inter1.left*_inter2.left;
      t=_inter1.right*_inter2.right;
      if (t>rr) rr=t;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (e)",ll,rr);
      #endif
    } else
    if (_inter1.left <=0) { // (-+)(++)
      ll=_inter1.left*_inter2.right;
      rr=_inter1.right*_inter2.right;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (f)",ll,rr);
      #endif
    } else // _inter1.left > 0
    if (_inter2.right <=0) { // (++)(--)
      ll=_inter1.right*_inter2.left;
      rr=_inter1.left*_inter2.right;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (g)",ll,rr);
      #endif
    } else // _inter2.right > 0
    if (_inter2.left <=0){ // (++)(-+)
      ll=_inter1.right*_inter2.left;
      rr=_inter1.right*_inter2.right;
      #ifdef DEBUGGING
      if(ll>rr)
        interval_error(" Operator * (h)",ll,rr);
      #endif
    }


    dec_double(&ll);
    #ifdef DEBUGGING
    if(ll>rr)
      interval_error(" Operator * (i)",ll,rr);
    #endif
    return interval(ll,rr);
}//--------------------------------------------------------------------

interval operator*(const interval& _inter, const bound_type _val)
{
   if((_inter.left==0) && (_inter.right==0)){
     return interval(0,0);
   }
   if(_val==0.0)      return interval(0,0);



   bound_type ll,rr;
   if(_val>=0)
   {
      ll = _inter.left*_val;
      dec_double(&ll);
      rr = _inter.right*_val;
   }
   else
   {
      ll = _inter.right*_val;
      dec_double(&ll);
      rr = _inter.left*_val;
   }
    #ifdef DEBUGGING
    if(ll>rr)
      interval_error(" Operator * (b)",ll,rr);
    #endif
      return interval(ll, rr);
}

interval operator*(const bound_type _val,const interval& _inter)
{
   bound_type ll,rr;
   if((_inter.left==0) && (_inter.right==0)){
     return interval(0,0);
   }
   if(_val==0.0)      return interval(0,0);

   if(_val>=0)
   {
      ll = _inter.left*_val;
      dec_double(&ll);
      rr = _inter.right*_val;
   }
   else
   {
      ll = _inter.right*_val;
      dec_double(&ll);
      rr = _inter.left*_val;
   }
    #ifdef DEBUGGING
    if(ll>rr)
      interval_error(" Operator * (c)",ll,rr);
    #endif
   return interval(ll,rr);
}//--------------------------------------------------------------------



//-------------------------------------------------
interval operator /(const interval& _inter1, const interval& _inter2)
{   bound_type ll,rr;
    if(_inter2.left<=0 && _inter2.right>=0){
      #if defined(KRAK)
        errorExit("Interval Error: division by zero");
      #else
       cerr << "Interval Error: division by zero.";
        exit(1);
      #endif
    }

   if(_inter1.right<=0 && _inter2.right<0)              // (-,-)/(-,-)
    {   ll=_inter1.right/_inter2.left;
        rr=_inter1.left/_inter2.right;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (--)(--)",ll,rr);
        #endif
        return( interval(ll,rr));
    }

   if(_inter1.right<=0 && _inter2.left>0)               // (-,-)/(+,+)
    {   ll=_inter1.left/_inter2.left;
        rr=_inter1.right/_inter2.right;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (--)(++)",ll,rr);
        #endif
        return( interval(ll,rr));
    }


    if(_inter1.left<=0 && _inter1.right>=0 && _inter2.right<0)  // (-,+)/(-,-)
    {   ll=_inter1.right/_inter2.right;
        rr=_inter1.left/_inter2.right;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (-+)(--)",ll,rr);
        #endif
        return( interval(ll,rr));
    }




    if(_inter1.left<=0 && _inter1.right>=0 && _inter2.left>0)   // (-,+)/(+,+)
    {   ll=_inter1.left/_inter2.left;
        rr=_inter1.right/_inter2.left;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (-+)(++)",ll,rr);
        #endif
        return( interval(ll,rr));
    }


    if(_inter1.left>=0 && _inter2.right<0)               // (+,+)/(-,-)
    {   ll=_inter1.right/_inter2.right;
        rr=_inter1.left/_inter2.left;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (++)(--)",ll,rr);
        #endif
        return( interval(ll,rr));
    }



    if(_inter1.left>=0 && _inter2.left>0)                // (+,+)/(+,+)
    {   ll=_inter1.left/_inter2.right;
        rr=_inter1.right/_inter2.left;
        dec_double(&ll);
        #ifdef DEBUGGING
        if(ll>rr)
          interval_error(" Operator / (++)(++)",ll,rr);
        #endif
        return( interval(ll,rr));
    }




  return(interval(0,0));
}//--------------------------------------------------------------------

//-----------------------------------------
interval operator %(const interval& _val1, const interval& _val2)
{

   if(_val2 <=0 )  // Dzielenie modulo przez zero lub liczbe ujemna
   {    cerr<<"Error *** Operator %% - domain error";
        exit(1);
   }

   interval temp = _val1;
   if( temp >= 0 )
     while( temp > _val2 )
       temp-=_val2;
   if( temp < 0 )
     while( temp < ( -1*_val2 ) )
       temp+=_val2;

   return temp;
}
interval operator %(const interval& _val1, const double _val2)
{
   if(_val2<=0)   // an attempt to modulo divide by  zero or negative number
   {    cerr<<"Error *** Operator %% - domain error";
        exit(1);
   }

   interval temp=_val1;

   if( temp>= 0 )
      while( temp > _val2 )
        temp-=_val2;
   if( temp < 0 )
      while( temp < -_val2 )
        temp+=_val2;

   return temp;
}//--------------------------------------------------------------------


//-------- sqrt -----------------
interval sqrt(const interval & z)
{
  interval retval;
  double temp;

  if(!(z>=0)){
    #if defined(KRAK)
      errorExit("Negative argument in sqrt(x)");
    #else
      cerr<<"Error *** Interval  Function sqrt(x) - domain error";
      exit(1);
    #endif
  }
  retval.right=sqrt(z.right);
  temp=sqrt(z.left);
  dec_double(&temp);
  if (temp <0) temp=0.0;
  retval.left=temp;


 return retval;
}//-------------------------------

void interval::split(interval &r){
  double xl,xr;
  xl=left;
  xr=right;
  left=right=(xl+xr)/2;
  r.left=left-xl;
  r.left=-r.left;
  r.right=xr-right;
  #ifdef DEBUGGING
  if(r.left>r.right)
      interval_error(" split",r.left,r.right);
  #endif
}


