#include "rounding.h"

#define ROUND_MASK    NOT 0C00h
#define ROUND_NEAREST 0
#define ROUND_DOWN    1 SHL 10
#define ROUND_UP      2 SHL 10
#define ROUND_CUT     3 SHL 10
#define IM_SET        0001h
#define IM_CLEAR      NOT 0001h

#if !defined(BORLAND) && !defined(SUN_OS)
//#if defined(LINUX) || defined(WINDOWS_GNU)
// LINUX code designed by Mikolaj Zalewski

  void round_nearest(void){
    short s;
    asm volatile (
      "fstcw %0\n\t"
      "andw $0xf3ff, %1\n\t"
      "orw $0x0000 , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }

  void round_down(void){
    short s;
    asm volatile (
      "fstcw %0\n\t"
      "andw $0xf3ff, %1\n\t"
      "orw $0x0400 , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }

  void round_up(void){
    short s;
    asm volatile (
      "fstcw %0\n\t"
      "andw $0xf3ff, %1\n\t"
      "orw $0x0800 , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }

  void round_cut(void){
    short s;
    asm volatile (
      "fstcw %0\n\t"
      "andw $0xf3ff, %1\n\t"
      "orw $0x0C00 , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }

  //needed under WINDOWS
  void set_intel_im(void){
    unsigned short s;
    asm volatile (
      "fstcw %0\n\t"
      "orw $0x0001 , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }

  void clear_intel_im(void){
    unsigned short s;
    asm volatile (
      "fstcw %0\n\t"
      "andw $0xfffe , %1\n\t"
      "fldcw %1"
      :"=m" (s)
      :"0" (s)
    );
  }


#elif defined(IBM)

void round_up(void){
  unsigned short fpu_up_cw;
                      asm              fstcw  fpu_up_cw;
                      asm              and    fpu_up_cw, ROUND_MASK;
                      asm              or     fpu_up_cw, ROUND_UP;
                      asm              fldcw  fpu_up_cw;
}

void round_down(void){
  unsigned short fpu_down_cw;
                      asm              fstcw  fpu_down_cw;
                      asm              and    fpu_down_cw, ROUND_MASK;
                      asm              or     fpu_down_cw, ROUND_DOWN;
                      asm              fldcw  fpu_down_cw;
}

void round_nearest(void){
  unsigned short fpu_nearest_cw;
                      asm              fstcw  fpu_nearest_cw;
                      asm              and    fpu_nearest_cw, ROUND_MASK;
                      asm              or     fpu_nearest_cw, ROUND_NEAREST;
                      asm              fldcw  fpu_nearest_cw;
}

void round_cut(void){
  unsigned short fpu_cut_cw;
                      asm              fstcw  fpu_cut_cw;
                      asm              and    fpu_cut_cw, ROUND_MASK;
                      asm              or     fpu_cut_cw, ROUND_CUT;
                      asm              fldcw  fpu_cut_cw;
}

void set_intel_im(void){
  unsigned short fpu_cut_cw;
                      asm              fstcw  fpu_cut_cw;
                      asm              or     fpu_cut_cw, IM_SET;
                      asm              fldcw  fpu_cut_cw;
}
void clear_intel_im(void){
  unsigned short fpu_cut_cw;
                      asm              fstcw  fpu_cut_cw;
                      asm              and    fpu_cut_cw, IM_CLEAR;
                      asm              fldcw  fpu_cut_cw;
}
#endif


int rounding_test(void){
  double a,b,nexta,preva,nextb,eps;
  int result=-1;

  nexta=a=1.5;
  do{
    preva=a;
    a=nexta;
    nexta=(a-1.0)/2+1.0;
  }while(a!=nexta);

  nextb=-1.5;
  do{
    b=nextb;
    nextb=(b+1.0)/2-1.0;
  }while(b!=nextb);

  eps=((2*(preva-1))/3+1);
  if(a == 1 && b == -1 && eps == 1){
    result=ROUNDING_MODE_CUT;
  }else{
    if(a>1  && (preva-1)/3>0){
      result=ROUNDING_MODE_UP;
    }else{
      if(b<-1){
        result=ROUNDING_MODE_DOWN;
      }else{
        result=ROUNDING_MODE_NEAREST;
      }
    }
  }

  return result;

}

void init_fpunit(void){
  #if(defined IBM) // this could also be implemented for LINUX and SPARC but so far was not needed
    // For some reason Win98 turns off some of the error handling of INTEL processors,
    // so we first turn on all numerical error handling
//    #ifndef WINDOWS_GNU          // Windows and Borland C++
    #if defined(BORLAND)          // Windows and Borland C++
       asm              fninit;
    #else                        // Windows and GNU C++
       asm volatile (
          "fninit\n\t"
      );
    #endif
    // next we turn on/off "forbidden numeric operation" error handling for INTEL processors
    #if(defined (INTEL_IM_CLEAR))
      clear_intel_im();
    #else
      set_intel_im();
    #endif
  #endif

  //finally we turn on rounding up, if necessary
  #ifdef UP_ROUNDING
    round_up();
  #endif
}
