(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     34174,        756]*)
(*NotebookOutlinePosition[     34869,        781]*)
(*  CellTagsIndexPosition[     34825,        777]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\
A semilinear fourth order elliptic problem with exponential \
nonlinearity\
\>", "Title",
  FontColor->GrayLevel[0.900008],
  Background->GrayLevel[0]],

Cell["G. Arioli, F. Gazzola, H.C. Grunau, E. Mitidieri", "Subsubtitle",
  Background->GrayLevel[0.900008]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["README",
  FontWeight->"Bold"]], "Subsubsection",
  FontSize->14,
  Background->GrayLevel[0.500008]],

Cell[TextData[{
  "This notebook can be used only together with the program \"gelfand\" which \
must be compiled separately from the C++ sources. The C++ sources,together \
with a makefile, are provided. To compile the program with g++ on a \
Unix/Linux machine simply create a directory, copy all files in the directory \
and type ",
  StyleBox["make",
    FontWeight->"Bold"],
  StyleBox[".\nIn the following line",
    FontVariations->{"CompatibilityType"->0}]
}], "Text"],

Cell[BoxData[
    \(\(link = Install["\<dir/gelfand\>"];\)\)], "Input",
  InitializationCell->True],

Cell[TextData[{
  "replace ",
  StyleBox["dir",
    FontSlant->"Italic"],
  " with the name of the directory where the file gelfand is stored. Run the \
\"Evaluation/Evaluate Initialization\" in the Kernel menu of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", and you are ready to go.\nIn section \"Definitions\" there are",
  " the definitions of all the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " commands used to interface with C++ and to set the initial data for the \
proof. In sections \"",
  Cell[BoxData[
      \(Estimates\ of\ \(\[Lambda]\^*\)\)],
    FontSize->14],
  "\" and \"",
  Cell[BoxData[
      RowBox[{"Estimates", " ", "of", 
        FormBox[\(\[Lambda]\_\[Sigma]\),
          "TraditionalForm"]}]],
    FontSize->14],
  "\" there are the commands used to prove Theorem 4 and to compute the \
values in table (10).\nTo compute the estimate for ",
  Cell[BoxData[
      \(\(\[Lambda]\^*\)\)],
    FontSize->14],
  " for, say, n=5 simply run the command\n",
  StyleBox["start[5]; meshstar[6.2829856, .001, 1000, 1]\n",
    FontWeight->"Bold"],
  StyleBox["The command ",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["start[5]",
    FontWeight->"Bold"],
  StyleBox[" initiates the values of the initial data to n=5, the command \
starts the computation of the lower bound of ",
    FontVariations->{"CompatibilityType"->0}],
  Cell[BoxData[
      \(\(\[Lambda]\^*\)\)],
    FontSize->14],
  StyleBox[" and produces, for each interval in the partition of the set \
\[Kappa], the rigorous estimate of the values of the fourth coordinate of the \
intersection with the hyperplane x1=4. When the command has computed the \
intersection for all elements of the partition it provides the final result. \
Note that, as discussed in the paper, only the lower bound of the resulting \
interval is meaningful.\nA note on the values of the parameter for the \
command ",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["meshstar",
    FontWeight->"Bold",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox[". The first is the angle of shooting, the second is the value of \
",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["r",
    FontWeight->"Bold",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox[" in the definition of the set \[Kappa], the third is the number \
of the intervals used in the partition, the fourth is 0 if no intermediate \
output is desider, otherwise 1.\n\nAnalogously, to compute the bounds on ",
    FontVariations->{"CompatibilityType"->0}],
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\_\[Sigma]\)],
    FontSize->14],
  " run the command\n",
  StyleBox["n=5;set;mesh [.001,20,1]",
    FontWeight->"Bold"],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Definitions",
  FontWeight->"Bold"]], "Subsubsection",
  FontSize->14,
  Background->GrayLevel[0.500008]],

Cell[BoxData[{
    RowBox[{
      RowBox[{\(GetTime[a_, n_] := 
            Block[{v1, v2, v3, v4, 
                T = 20}, {v1, v2, v3, v4} = {v1, v2, v3, v4} /. 
                  First[NDSolve[{\(v1'\)[t] \[Equal] \((2 - n)\) v1[t] - 
                            v2[t] + 4 \((n - 2)\), \(v2'\)[t] \[Equal] 
                          2  v2[t] + v3[t], \(v3'\)[
                            t] \[Equal] \((4 - n)\) v3[t] + v4[t], \(v4'\)[
                            t] \[Equal] v1[t] v4[t], 
                        v1[0] \[Equal] Averint[a[\([1]\)]], 
                        v2[0] \[Equal] Averint[a[\([2]\)]], 
                        v3[0] \[Equal] Averint[a[\([3]\)]], 
                        v4[0] \[Equal] Averint[a[\([4]\)]]}, {v1, v2, v3, 
                        v4}, {t, 0, T}, 
                      StoppingTest \[Rule] {v1[t] \[GreaterEqual] 4}, 
                      MaxSteps \[Rule] 30000]]; \[IndentingNewLine]T = 
                v1[\([1, 1, 2]\)]; \[IndentingNewLine]t /. 
                FindRoot[v1[t] \[Equal] 4. , {t, {T,  .95  T}}]];\), " ", 
        StyleBox[\( (*\ 
            this\ function\ estimates\ the\ time\ from\ the\ starting\ point\ \
close\ to\ P2\ to\ the\ hyperplane\ v1 = 4\ *) \),
          "Text",
          FontFamily->"Times",
          FontWeight->"Plain",
          FontVariations->{"CompatibilityType"->0}]}], 
      StyleBox["\[IndentingNewLine]",
        FontFamily->"Times"]}], "\n", 
    RowBox[{
      RowBox[{\(mat[
              n_] := {{2 - n, \(-1\), 0, 0}, \[IndentingNewLine]{0, 2, 1, 
                0}, \[IndentingNewLine]{0, 0, \((4 - n)\), 
                1}, \[IndentingNewLine]{\(-8\) \((n - 2)\) \((n - 4)\), 0, 0, 
                0}};\), " ", 
        StyleBox[\( (*\ the\ matrix\ A2\ *) \),
          FontFamily->"Times",
          FontWeight->"Plain",
          FontVariations->{"CompatibilityType"->0}]}], 
      "\[IndentingNewLine]"}], "\n", 
    RowBox[{
      RowBox[{\(ev[n_] := 
            Block[{ev, evalues, evectors, i, modu}, 
              evalues = 
                Eigenvalues[mat[1. \ n]]; \[IndentingNewLine]evectors = 
                Eigenvectors[mat[1. \ n]]; \[IndentingNewLine]Do[
                If[Re[evalues[\([i]\)]] > 0, evector = evectors[\([i]\)], 
                  Null], {i, 4}]; \[IndentingNewLine]modu = 
                Sqrt[evector . evector]; \[IndentingNewLine]evector/modu];\), 
        " ", 
        StyleBox[\( (*\ 
            conputation\ of\ the\ normalized\ eigenvalues\ of\ A2\ *) \),
          FontFamily->"Times",
          FontWeight->"Plain",
          FontVariations->{"CompatibilityType"->0}]}], 
      "\[IndentingNewLine]"}], "\n", 
    RowBox[{
      RowBox[{\(mat2[
              n_] := {{2 - n, \(-1\), 0, 0}, \[IndentingNewLine]{0, 2, 1, 
                0}, \[IndentingNewLine]{0, 0, \((4 - n)\), 
                1}, \[IndentingNewLine]{0, 0, 0, 4}};\), " ", 
        StyleBox[\( (*\ the\ matrix\ A1\ *) \),
          FontFamily->"Times",
          FontWeight->"Plain",
          FontVariations->{"CompatibilityType"->0}]}], 
      "\[IndentingNewLine]"}], "\n", 
    RowBox[{\(GetTime2[a_] := 
          Block[{v1, v2, v3, v4, T = 20}, \[IndentingNewLine]If[
              Averint[a[\([1]\)]] > 4, sign = \(-1. \), 
              sign = 1. ]; \[IndentingNewLine]{v1, v2, v3, 
                v4} = {v1, v2, v3, v4} /. 
                First[NDSolve[{\(v1'\)[t] \[Equal] \((2 - n)\) v1[t] - 
                          v2[t] + 4 \((n - 2)\), \[IndentingNewLine]\(v2'\)[
                          t] \[Equal] 
                        2  v2[t] + v3[t], \[IndentingNewLine]\(v3'\)[
                          t] \[Equal] \((4 - n)\) v3[t] + 
                          v4[t], \[IndentingNewLine]\(v4'\)[t] \[Equal] 
                        v1[t] v4[t], \[IndentingNewLine]v1[0] \[Equal] 
                        Averint[a[\([1]\)]], 
                      v2[0] \[Equal] 
                        Averint[a[\([2]\)]], \[IndentingNewLine]v3[
                          0] \[Equal] Averint[a[\([3]\)]], 
                      v4[0] \[Equal] 
                        Averint[a[\([4]\)]]}, \[IndentingNewLine]{v1, v2, v3, 
                      v4}, {t, 0, T}, 
                    StoppingTest \[Rule] {sign\ v1[t] \[GreaterEqual] 
                          sign\ 4}, 
                    MaxSteps \[Rule] 30000]]; \[IndentingNewLine]T = 
              v1[\([1, 1, 2]\)]; \[IndentingNewLine]t /. 
              FindRoot[v1[t] \[Equal] 4. , {t, {T,  .95  T}}]];\), 
      StyleBox[\( (*\ 
          this\ function\ estimates\ the\ time\ from\ the\ starting\ point\
\[IndentingNewLine]close\ to\ \(P  1\)\ to\ the\ hyperplane\ v1 = 4\ *) \),
        "Text",
        FontFamily->"Times",
        FontWeight->"Plain",
        FontVariations->{"CompatibilityType"->0}]}]}], "Input",
  InitializationCell->True],

Cell[BoxData[{
    \(\(SetDirectory["\<~/math/gelfand\>"];\)\), "\[IndentingNewLine]", 
    \(\(d = 4;\)\), "\[IndentingNewLine]", 
    \(\(f[x_] := {\((2 - n)\) x[\([1]\)] - x[\([2]\)] + 4 \((n - 2)\), 
          2  x[\([2]\)] + 
            x[\([3]\)], \[IndentingNewLine]\((4 - n)\) x[\([3]\)] + 
            x[\([4]\)], x[\([1]\)]\ x[\([4]\)]};\)\), "\[IndentingNewLine]", 
    \(\(dftemp = {D[f[{x, v, y, w}], x], D[f[{x, v, y, w}], v], 
          D[f[{x, v, y, w}], y], 
          D[f[{x, v, y, w}], w]};\)\), "\[IndentingNewLine]", 
    \(df2 = {D[dftemp, x], D[dftemp, v], D[dftemp, y], D[dftemp, w]}; 
    df[a_] := Block\ [{x, v, y, w, 
          e}, \[IndentingNewLine]dftemp /. {x \[Rule] a[\([1]\)], 
            w \[Rule] 
              a[\([4]\)]}];\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(Enclosure[x_, t_] := 
        Block[{i, x00 = x, y, z, zeroh = Interval[{0. , t}]}, 
          z = x00 + zeroh\ f[x00]\  + 
              Table[Interval[{\(- .0001\),  .0001}], {i, 
                  4}]; \[IndentingNewLine]y = 
            x00 + zeroh\ f[z]; \[IndentingNewLine]While[\(! incl[z, 
                y]\), \[IndentingNewLine]Do[\[IndentingNewLine]If[\(! \
IntervalMemberQ[z[\([i]\)], y[\([i]\)]]\), 
                z[\([i]\)] = z[\([i]\)] + Interval[{\(- .005\),  .005}], 
                Null]\  (*\ end\ if\ *) \[IndentingNewLine], {i, 4}]; \  (*\ 
              end\ do\ *) \[IndentingNewLine]y = 
              x00 + zeroh\ f[z]\[IndentingNewLine]]; \  (*\ 
            end\ while\ *) \[IndentingNewLine]\ 
          y];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(incl[z_, y_] := 
        Block[{inc = True, 
            vectbool = IntervalMemberQ[z, y]}, \[IndentingNewLine]Do[
            inc = inc && vectbool[\([i]\)], {i, d}]; 
          inc];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(Averint[a_] :=  .5 \((Max[a] + Min[a])\);\)\)}], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(MappaT[v_, T_] := 
      Block[{v1, v2, v3, 
          v4}, \ \[IndentingNewLine]{v1, v2, v3, v4} = {v1, v2, v3, v4} /. 
            First[NDSolve[\[IndentingNewLine]{\(v1'\)[
                      t] == \((2 - n)\) v1[t] - v2[t] + 
                      4 \((n - 2)\), \[IndentingNewLine]\(v2'\)[t] \[Equal] 
                    2  v2[t] + v3[t], \[IndentingNewLine]\(v3'\)[
                      t] \[Equal] \((4 - n)\) v3[t] + 
                      v4[t], \[IndentingNewLine]\(v4'\)[t] \[Equal] 
                    v1[t] v4[t], \[IndentingNewLine]v1[0] == v[\([1]\)], 
                  v2[0] \[Equal] v[\([2]\)], v3[0] == v[\([3]\)], 
                  v4[0] \[Equal] v[\([4]\)]}, \[IndentingNewLine]{v1, v2, v3, 
                  v4}, {t, 0, T}, 
                StoppingTest \[Rule] {v1[t] \[GreaterEqual] 4}, 
                MaxSteps \[Rule] 30000]]; \[IndentingNewLine]{v1[T], v2[T], 
          v3[T], v4[T]}]\)], "Input"],

Cell[BoxData[{
    \(\(TestInt[l_, i_, div_] := 
        Block[{a, b, c, d, j, p}, \[IndentingNewLine]p = Length[l]; 
          j = If[i < p, i + 1, 1]; \[IndentingNewLine]a = l[\([i]\)]; 
          b = l[\([i]\)] + 
              div \((l[\([j]\)] - l[\([i]\)])\); \[IndentingNewLine]c = 
            Interval[{a[\([1]\)], b[\([1]\)]}]; 
          d = Interval[{a[\([2]\)], b[\([2]\)]}]; {c, d}];\)\  (*\ 
      extracts\ interval\ from\ t - set\ *) \), "\[IndentingNewLine]", 
    \(\(d = 4;\)\), "\[IndentingNewLine]", 
    \(\(Error2[y_, order_, 
          t_] := \[IndentingNewLine]Block[{e1, e2, e3, 
            o = order + 1}, \[IndentingNewLine]e1 = 
            Int2Array[Enclosure2[y, t]]; \[IndentingNewLine]e2 = 
            error[e1, o, n, t]; \[IndentingNewLine]e3 = 
            Array2Int[
              e2]; \[IndentingNewLine]t\^o\ e3\[IndentingNewLine]];\)\), "\
\[IndentingNewLine]", 
    \(\(\[CapitalPhi]2[y_, order_, t_] := 
        Block[{e1, e2}, \[IndentingNewLine]e1 = 
            Int2Array[y]; \[IndentingNewLine]e2 = 
            Phi[e1, order, n, t]; \[IndentingNewLine]Array2Int[
            e2]\[IndentingNewLine]];\)\), "\[IndentingNewLine]", 
    \(\(d\[CapitalPhi]2[y_, order_, t_] := 
        Block[{e1, e2}, \[IndentingNewLine]e1 = 
            Int2Array[y]; \[IndentingNewLine]e2 = 
            dPhi[e1, order, n, t]; \[IndentingNewLine]Array2Mat[
            e2]\[IndentingNewLine]];\)\), "\[IndentingNewLine]", 
    \(\(Enclosure2[y_, t_] := 
        Array2Int[Encl[Int2Array[y], n, t]];\)\), "\[IndentingNewLine]", 
    \(\(Int2Array[x_] := {Min[x[\([1]\)]], Max[x[\([1]\)]], Min[x[\([2]\)]], 
          Max[x[\([2]\)]], Min[x[\([3]\)]], Max[x[\([3]\)]], Min[x[\([4]\)]], 
          Max[x[\([4]\)]]};\)\), "\[IndentingNewLine]", 
    \(\(Mat2Array[x_] := 
        Block[{e1, e2}, \[IndentingNewLine]e1 = 
            Flatten[x]; \[IndentingNewLine]e2 = 
            Table[{Min[e1[\([i]\)]], Max[e1[\([i]\)]]}, {i, 
                16}]; \[IndentingNewLine]Flatten[
            e2]\[IndentingNewLine]];\)\), "\[IndentingNewLine]", 
    \(\(Array2Int[x_] := {Interval[{x[\([1]\)], x[\([2]\)]}], 
          Interval[{x[\([3]\)], x[\([4]\)]}], 
          Interval[{x[\([5]\)], x[\([6]\)]}], 
          Interval[{x[\([7]\)], x[\([8]\)]}]};\)\), "\[IndentingNewLine]", 
    \(\(Array2Mat[
          x_] := {{Interval[{x[\([1]\)], x[\([2]\)]}], 
            Interval[{x[\([3]\)], x[\([4]\)]}], 
            Interval[{x[\([5]\)], x[\([6]\)]}], 
            Interval[{x[\([7]\)], 
                x[\([8]\)]}]}, \[IndentingNewLine]{Interval[{x[\([9]\)], 
                x[\([10]\)]}], Interval[{x[\([11]\)], x[\([12]\)]}], 
            Interval[{x[\([13]\)], x[\([14]\)]}], 
            Interval[{x[\([15]\)], 
                x[\([16]\)]}]}, \[IndentingNewLine]{Interval[{x[\([17]\)], 
                x[\([18]\)]}], Interval[{x[\([19]\)], x[\([20]\)]}], 
            Interval[{x[\([21]\)], x[\([22]\)]}], 
            Interval[{x[\([23]\)], 
                x[\([24]\)]}]}, \[IndentingNewLine]{Interval[{x[\([25]\)], 
                x[\([25]\)]}], Interval[{x[\([27]\)], x[\([28]\)]}], 
            Interval[{x[\([29]\)], x[\([30]\)]}], 
            Interval[{x[\([31]\)], 
                x[\([32]\)]}]}};\)\), "\[IndentingNewLine]", 
    \(pos[a_] := Min[a] \[GreaterEqual] 4; neg[a_] := Max[a] \[LessEqual] 4; 
    WidthInt[a_] := Max[a] - Min[a];\), "\[IndentingNewLine]", 
    \(\(hull[a_, b_] := 
        Interval[{Min[a, b], Max[a, b]}];\)\), "\[IndentingNewLine]", 
    \(\(id = Interval[0. ] + IdentityMatrix[4];\)\), "\[IndentingNewLine]", 
    \(\(go[t_] := 
        Block[{order = 12, i, j, temp, w}, \[IndentingNewLine]w = 
            Error2[x00, order, 
              t]; \[IndentingNewLine]temp = \[CapitalPhi]2[xreal, order, t] + 
              w; \[IndentingNewLine]xrealold = xreal; x00old = x00; 
          rtilde2old = rtilde2; \[IndentingNewLine]newCold = newC; 
          Cmold = Cm; 
          Bold = B; \[IndentingNewLine]xreal = 
            Table[Interval[Averint[temp[\([i]\)]]], {i, 
                d}]; \[IndentingNewLine]z = 
            temp - xreal; \[IndentingNewLine]Areal = 
            d\[CapitalPhi]2[x00, order, t]; \[IndentingNewLine]AC = 
            Areal . Cm; \[IndentingNewLine]newC = 
            Table[Averint[AC[\([i, j]\)]], {i, d}, {j, 
                d}]; \[IndentingNewLine]AB2 = 
            Areal . B; \[IndentingNewLine]AB = 
            Table[Interval[Averint[AB2[\([i, j]\)]]], {i, d}, {j, 
                d}]; \[IndentingNewLine]{Bt, RR} = 
            QRDecomposition[AB]; \[IndentingNewLine]B = 
            Transpose[Bt]; \[IndentingNewLine]rtilde = 
            AB2 . rtilde2 + 
              z + \((AC - newC)\) . 
                r0; \[IndentingNewLine]rtilde2 = \((Bt . AB2)\) . rtilde2 + 
              Bt . z; \[IndentingNewLine]r = 
            newC . r0 + rtilde; \[IndentingNewLine]Cm = 
            newC; \[IndentingNewLine]x00 = xreal + r];\)\)}], "Input",
  InitializationCell->True],

Cell[BoxData[{
    \(\(Section[s_] := 
        Block[{ch1, ch2, bef}, \[IndentingNewLine]zeroh = 
            Interval[{0. , s}]; \[IndentingNewLine]bef = 
            x00[\([4]\)]; \[IndentingNewLine]ch1 = 
            x00[\([1]\)] - 4; \[IndentingNewLine]go[s]; \[IndentingNewLine]If[
            star, CheckDirectionStar, 
            CheckDirection]; \[IndentingNewLine]ch2 = 
            x00[\([1]\)] - 4; \[IndentingNewLine]If[
            Max[ch1\ ch2] \[GreaterEqual] 0, 
            If[P \[Equal] 1, Print["\<Error Poincar\[EAcute]\>"], 
              Save["\<error.mx\>", i, j, n, ch1, ch2]], 
            Null]; \[IndentingNewLine]bef = \(-hull[bef, 
                x00[\([4]\)]]\); \
\[IndentingNewLine]bef];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(MainStar[a_, 
          tt_] := \[IndentingNewLine]Block[{}, \[IndentingNewLine]x00 = a; \ 
          t = tt; \[IndentingNewLine]zeroh = 
            Interval[{0. , t}]; \[IndentingNewLine]rtilde2 = 
            Table[Interval[0. ], {i, 4}]; \[IndentingNewLine]rtilde = 
            Table[Interval[0. ], {i, 4}]; \[IndentingNewLine]B = 
            IdentityMatrix[4]; \[IndentingNewLine]Bt = 
            B; \[IndentingNewLine]xreal = 
            Table[Interval[Averint[x00[\([i]\)]]], {i, 
                4}]; \[IndentingNewLine]r0 = 
            x00 - xreal; \[IndentingNewLine]Cm = 
            IdentityMatrix[d]; \[IndentingNewLine]While[
            Max[x00[\([1]\)]] < 4, CheckDirectionStar; 
            go[t]]; \[IndentingNewLine]x00 = x00old; xreal = xrealold; 
          rtilde2 = rtilde2old; \[IndentingNewLine]newC = newCold; 
          Cm = Cmold; 
          B = Bold; \[IndentingNewLine]t =  .0001; \[IndentingNewLine]While[
            Max[x00[\([1]\)]] < 4, CheckDirectionStar; 
            go[t]]; \[IndentingNewLine]x00 = x00old; xreal = xrealold; 
          rtilde2 = rtilde2old; \[IndentingNewLine]newC = newCold; 
          Cm = Cmold; 
          B = Bold; \[IndentingNewLine]t = 
            1.1  Abs[
                Averint[x00[\([1]\)] - 4]/
                  Averint[\(f[x00]\)[\([1]\)]]]; \[IndentingNewLine]star = 
            True; \[IndentingNewLine]Section[
            t]\[IndentingNewLine]];\)\[IndentingNewLine]\), "\
\[IndentingNewLine]", 
    \(\(CheckDirection := 
        Block[{x = Enclosure2[x00, t]}, \[IndentingNewLine]If[
            4 \((n - 2)\) < Max[\((n - 2)\) x[\([1]\)] + x[\([2]\)]], 
            If[P \[Equal] 1, Print["\<Error direction\>"], 
              Save["\<error.mx\>", i, j, n]], 
            Null]];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(CheckDirectionStar := 
        Block[{x = Enclosure2[x00, t]}, \[IndentingNewLine]If[
            Min[x[\([1]\)]\ x[\([4]\)] \((x[\([1]\)] - 
                        4)\) - \((\((2 - n)\) x[\([1]\)] - x[\([2]\)] + 
                        4 \((n - 2)\))\) \((x[\([4]\)] + 150)\)] \[LessEqual] 
              0. , \[IndentingNewLine]If[P \[Equal] 1, 
              Print["\<Error direction\>"], Save["\<error.mx\>", i, j, n]], 
            Null]];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(Main[a_, 
          div_] := \[IndentingNewLine]Block[{x00 = a, 
            t}, \[IndentingNewLine]t =  .995 
               GetTime[a, n]/div; \[IndentingNewLine]zeroh = 
            Interval[{0, t}]; \[IndentingNewLine]rtilde2 = 
            Table[Interval[0. ], {i, 4}]; \[IndentingNewLine]rtilde = 
            rtilde2; \[IndentingNewLine]B = 
            IdentityMatrix[4]; \[IndentingNewLine]Bt = 
            B; \[IndentingNewLine]xreal = 
            Table[Interval[Averint[x00[\([i]\)]]], {i, 
                4}]; \[IndentingNewLine]r0 = 
            x00 - xreal; \[IndentingNewLine]Cm = 
            IdentityMatrix[d]; \[IndentingNewLine]Do[CheckDirection; 
            go[t], {i, 
              div}]; \[IndentingNewLine]t =  .0001; \[IndentingNewLine]While[
            Max[x00[\([1]\)]] < 4, CheckDirection; 
            go[t]]; \[IndentingNewLine]x00 = x00old; xreal = xrealold; 
          rtilde2 = rtilde2old; \[IndentingNewLine]newC = newCold; 
          Cm = Cmold; 
          B = Bold; \[IndentingNewLine]t = 
            1.1  Abs[
                Averint[x00[\([1]\)] - 4]/
                  Averint[\(f[x00]\)[\([1]\)]]]; \[IndentingNewLine]star = 
            False; \[IndentingNewLine]Section[
            t]\[IndentingNewLine]];\)\[IndentingNewLine]\), "\
\[IndentingNewLine]", 
    \(\(MainHigh[
          a_] := \[IndentingNewLine]Block[{x00 = a, 
            t =  .01}, \[IndentingNewLine]zeroh = 
            Interval[{0, t}]; \[IndentingNewLine]rtilde2 = 
            Table[Interval[0. ], {i, 4}]; \[IndentingNewLine]rtilde = 
            rtilde2; \[IndentingNewLine]B = 
            IdentityMatrix[4]; \[IndentingNewLine]Bt = 
            B; \[IndentingNewLine]xreal = 
            Table[Interval[Averint[x00[\([i]\)]]], {i, 
                4}]; \[IndentingNewLine]r0 = 
            x00 - xreal; \[IndentingNewLine]Cm = 
            IdentityMatrix[d]; \[IndentingNewLine]While[
            Max[x00[\([1]\)]] < 4, CheckDirection; 
            go[t]]; \[IndentingNewLine]x00 = x00old; xreal = xrealold; 
          rtilde2 = rtilde2old; \[IndentingNewLine]newC = newCold; 
          Cm = Cmold; B = Bold; \[IndentingNewLine]t =  .0001; 
          zeroh = Interval[{0, t}]; \[IndentingNewLine]While[
            Max[x00[\([1]\)]] < 4, CheckDirection; 
            go[t]]; \[IndentingNewLine]x00 = x00old; xreal = xrealold; 
          rtilde2 = rtilde2old; \[IndentingNewLine]newC = newCold; 
          Cm = Cmold; 
          B = Bold; \[IndentingNewLine]t = 
            1.2  Abs[
                Averint[x00[\([1]\)] - 4]/
                  Averint[\(f[x00]\)[\([1]\)]]]; \[IndentingNewLine]star = 
            False; \[IndentingNewLine]Section[
            t]\[IndentingNewLine]];\)\)}], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(End\)], "Input",
  Background->GrayLevel[0.900008]]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Estimates\ of\ \(\[Lambda]\^*\)\)], "Subsection",
  FontSize->14,
  Background->GrayLevel[0.666667]],

Cell[BoxData[{
    \(\(start[a_] := 
        Block[{Ab, Mb, M2b}, \[IndentingNewLine]n = 
            a; \[IndentingNewLine]Ab = {{2 - n, \(-1\), 0, 0}, {0, 2, 1, 
                0}, {0, 0, \((4 - n)\), 1}, {0, 0, 0, 4}}; 
          Mb = Transpose[Eigenvectors[Ab]]; 
          M2b = Transpose[Inverse[Mb]]; \[IndentingNewLine]Bb = 
            Simplify[Inverse[Mb] . Ab . Mb]; \[IndentingNewLine]\[Alpha] = 
            4 - n; \[Alpha]p = 
            2; \[IndentingNewLine]\[Gamma] =  .5 \((n - 4)\)/
                Sqrt[Sum[\ Mb[\([4, i]\)]\^2\ , {i, 4}] 
                    Sum[Mb[\([1, i]\)]\^2, {i, 4}] 
                    Sum[M2b[\([4, i]\)]\^2, {i, 4}]]; \[IndentingNewLine]rho[
              r_] :=  .5 \((\[Gamma] - 
                  Sqrt[\[Gamma]\^2 - 4  r\^2])\); \[IndentingNewLine]k[
              r_] := \(\[Gamma]\^2 - \@\(\(-4\)\ r\^2\ \[Gamma]\^2 + \[Gamma]\
\^4\)\)\/\(2\ r\^2\); \[IndentingNewLine]\[Delta]1[
              r_] := \[Alpha]p + \[Alpha]/\[Gamma] \((r + \(\(k[\[Rho]]\^2\) \
r\^3\)\/\[Gamma]\^2)\); \[IndentingNewLine]\[Delta]2[
              r_] := \((k[r]/\[Gamma])\)^2 + 1 - 
              k[r]];\)\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(meshstar[\[Phi]_, r_, maxi_, P_] := 
        Block[{A, M, i, k, d\[Theta], \[Rho] = rho[r], res, out = 0, v1, v2, 
            x1, y1, x2, y2, st1 = r\ Cos[\[Phi]], 
            st2 = r\ Sin[\[Phi]]}, \[IndentingNewLine]If[\((\[Delta]1[
                    r] \[LessEqual] 0)\) || \((\[Delta]2[r] \[LessEqual] 
                  0)\), Print["\<Error\>"], 
            Null]; \[IndentingNewLine]d\[Theta] = 
            2  \[Pi]/maxi; \[IndentingNewLine]A = {{2 - n, \(-1\), 0, 
                0}, \[IndentingNewLine]{0, 2, 1, 0}, \[IndentingNewLine]{0, 
                0, \((4 - n)\), 1}, \[IndentingNewLine]{0, 0, 0, 
                4}}; \[IndentingNewLine]M = 
            Transpose[
              Eigenvectors[
                A]]; \[IndentingNewLine]Do[\[IndentingNewLine]x1 = \[Rho]\ \
Cos[i\ d\[Theta]]; 
            x2 = \[Rho]\ Cos[
                  i\ d\[Theta] + 
                    d\[Theta]]; \[IndentingNewLine]y1 = \[Rho]\ Sin[
                  i\ d\[Theta]]; 
            y2 = \[Rho]\ Sin[
                  i\ d\[Theta] + \ d\[Theta]]; \[IndentingNewLine]v1 = {4, 0, 
                  0, 0} + M . {x1, y1, st1, st2}; 
            v2 = {4, 0, 0, 0} + 
                M . {x2, y2, st1, st2}; \[IndentingNewLine]a = 
              Table[Interval[{v1[\([k]\)], v2[\([k]\)]}], {k, 1, 
                  4}]; \[IndentingNewLine]res = 
              MainStar[a,  .02]; \[IndentingNewLine]If\ [out \[Equal] 0, 
              out = res, Null]; \[IndentingNewLine]out = 
              hull[out, res]; \[IndentingNewLine]If[Mod[i, 1] \[Equal] 0, 
              If[P \[Equal] 1, Print[out], Save["\<gs.mx\>", i, out]], 
              Null], \[IndentingNewLine]{i, 0, 
              maxi - 1}]; \[IndentingNewLine]If[P \[Equal] 1, Print[out], 
            Save["\<gelfstar.mx\>", n, 
              out]];\[IndentingNewLine]];\)\)}], "Input",
  InitializationCell->True],

Cell[BoxData[
    \(The\ following\ commands\ make\ the\ complete\ computation\ of\ the\ \
lower\ estimate\ for\ \(\[Lambda]\^*\)\ for\ n = 5,  ... , 10\)], "Text",
  FontFamily->"Times"],

Cell[BoxData[
    \(start[5]; meshstar[6.2829856,  .001, 1000, 1]\)], "Input"],

Cell[BoxData[
    \(start[6]; meshstar[6.28298854,  .001, 1000, 1]\)], "Input"],

Cell[BoxData[
    \(start[7]; meshstar[6.2829901,  .001, 1000, 1]\)], "Input"],

Cell[BoxData[
    \(start[8]; meshstar[6.2829918,  .001, 1000, 1]\)], "Input"],

Cell[BoxData[
    \(start[9]; meshstar[6.2829914,  .001, 1000, 1]\)], "Input"],

Cell[BoxData[
    \(start[10]; meshstar[6.28316589,  .0001, 10000, 1]\)], "Input"],

Cell[BoxData[
    \(End\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    RowBox[{"Estimates", " ", "of", 
      FormBox[\(\[Lambda]\_\[Sigma]\),
        "TraditionalForm"]}]], "Subsection",
  FontSize->14,
  Background->GrayLevel[0.666667]],

Cell[BoxData[{
    \(\(set := {A = mat[n]; 
          evA = Eigenvectors[A]; \[IndentingNewLine]If[
            n \[LessEqual] 12, \[IndentingNewLine]temp1 = 
              Simplify[\((evA[\([1]\)] + evA[\([2]\)])\)/
                  2]; \[IndentingNewLine]temp2 = 
              Simplify[\((evA[\([1]\)] - evA[\([2]\)])\)/2  
                  I]; \[IndentingNewLine]evect = {temp1, temp2, 
                Simplify[evA[\([3]\)]], 
                Simplify[
                  evA[\([4]\)]]}; \[IndentingNewLine]\[Alpha] = \((4 - n)\)/
                2, \[IndentingNewLine]evect = {Simplify[evA[\([1]\)]], 
                Simplify[evA[\([2]\)]], Simplify[evA[\([3]\)]], 
                Simplify[evA[\([4]\)]]}; \[IndentingNewLine]temp1 = 
              Eigenvalues[A]; \[Alpha] = temp1[\([1]\)]; 
            Do[If[\((temp1[\([i]\)] > \[Alpha])\) && \((temp1[\([i]\)] < 
                      0)\), \[Alpha] = temp1[\([i]\)], Null], {i, 2, 
                4}]]; \  (*\ 
            endIf\ *) \[IndentingNewLine]\[IndentingNewLine]\[Alpha]p = 
            1\/2\ \((4 - 
                  n + \@\(8 - 4\ n + n\^2 + 4\ \@\(68 - 52\ n + 9\ \
n\^2\)\))\); \[IndentingNewLine]M = Transpose[evect]; \[IndentingNewLine]M2 = 
            Transpose[Inverse[M]]; \[IndentingNewLine]B = 
            Simplify[Inverse[M] . A . M]; \[IndentingNewLine]change[
              x_] := {0, 4  n - 8, 
                16 - 8  n, \(-8\) \((n - 2)\) \((n - 4)\)} + 
              M . x; \[IndentingNewLine]\[Gamma] = \
\(-\[Alpha]\)/\[Sqrt]\((Sum[\ M[\([4, i]\)]\^2\ \ , {i, 4}] 
                    Sum[M[\([1, i]\)]\^2, {i, 4}] 
                    Sum[M2[\([4, i]\)]\^2, {i, 
                        4}])\); \[IndentingNewLine]rad[\[Rho]_] := \[Rho]\ \
Sqrt[\[Gamma]/\[Rho] - 1]; 
          rho[r_] :=  .5 \((\[Gamma] - \@\(\[Gamma]\^2 - \(\(4\)\(\ \
\)\(r\^2\)\(\ \)\)\))\); \[IndentingNewLine]k[\[Rho]_] := \(\[Gamma]\^2 - \
\@\(\(-4\)\ rad[\[Rho]]\^2\ \[Gamma]\^2 + \[Gamma]\^4\)\)\/\(2\ \
rad[\[Rho]]\^2\); \[IndentingNewLine]\[Delta][\[Rho]_] := \[Alpha]p + \
\[Alpha]/\[Gamma] \((rad[\[Rho]] + \(\(k[\[Rho]]\^2\) rad[\[Rho]]\^3\)\/\
\[Gamma]\^2)\);\[IndentingNewLine]};\)\[IndentingNewLine]\), "\
\[IndentingNewLine]", 
    \(\(mesh[R_, M_, P_] := 
        Block[{i, j, k, dz, d\[Theta], maxi = M, res, 
            out = 0, \[IndentingNewLine]v, v1, v2, x1, y1, z1, x2, y2, z2, 
            r, \[Rho]}, \[IndentingNewLine]rr = 
            Interval[R]; \[IndentingNewLine]\[Rho] = 
            rho[rr]; \[IndentingNewLine]If[\((Min[\[Delta][\[Rho]]] <= 
                  0. )\) || \((Min[k[\[Rho]]] >= 
                  Max[\((k[\[Rho]]/\[Gamma])\)^2 + 1])\), 
            Save["\<error.mx\>", n]; Print["\<Error\>"], 
            Null]; \[IndentingNewLine]dz = 
            2. /maxi; \[IndentingNewLine]Save["\<g.mx\>", 
            n]; \[IndentingNewLine]Do[z1 = \[Rho]\ \((\(-1\) + i\ dz)\); 
            z2 = \[Rho]\ \((\(-1\) + \((i + 1)\) dz)\); 
            maxj = Floor[ .5  
                  maxi \@\( 1.1 - \((\(2  i + 1\)\/maxi - 1)\)\^2\)]; 
            d\[Theta] = 
              2  \[Pi]/
                  maxj; \[IndentingNewLine]Do[\[IndentingNewLine]x1 = \[Rho]\ \
\(\@\(1 - \((\(-1\) + i\ dz)\)\^2\)\) Cos[j\ d\[Theta]]; 
              y1 = \[Rho]\ \(\@\(1 - \((\(-1\) + i\ dz)\)\^2\)\) 
                  Sin[j\ d\[Theta]]; \[IndentingNewLine]x2 = \[Rho]\ \(\@\(1 \
- \((\(-1\) + \((i + 1)\) dz)\)\^2\)\) 
                  Cos[j\ d\[Theta] + \ 
                      d\[Theta]]; \[IndentingNewLine]y2 = \[Rho]\ \(\@\(1 - \
\((\(-1\) + \((i + 1)\) dz)\)\^2\)\) 
                  Sin[j\ d\[Theta] + \ d\[Theta]]; \[IndentingNewLine]v1 = 
                change[{x1, y1, z1, \(-rr\)}]; 
              v2 = change[{x2, y2, z2, \(-rr\)}]; \[IndentingNewLine]a = 
                Table[hull[v1[\([k]\)], v2[\([k]\)]], {k, 1, 
                    4}]; \[IndentingNewLine]If[n <= 10, res = Main[a, 100], 
                res = MainHigh[a]]; \[IndentingNewLine]If[P \[Equal] 1, 
                Print[res], Null]; \[IndentingNewLine]If\ [out == 0, 
                out = res, out = hull[out, res]], {j, 0, 
                maxj}]; \[IndentingNewLine]Save["\<g.mx\>", i, n, out], {i, 
              0, maxi - 1}]; \[IndentingNewLine]r = 
            R; \[IndentingNewLine]Save["\<gelfand.mx\>", n, r, maxi, 
            out]\[IndentingNewLine]];\)\)}], "Input",
  InitializationCell->True],

Cell[BoxData[{\(The\ following\ commands\ make\ the\ complete\ computation\ \
of\ the\), "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"estimates", " ", "for", " ", 
          FormBox[\(\[Lambda]\_\[Sigma]\),
            "TraditionalForm"], " ", "for", " ", "n"}], "=", "5"}], ",", "...",
       ",", "16"}]}], "Text",
  FontFamily->"Times"],

Cell[BoxData[
    \(n = 5; set; mesh\ [ .001, 20, 1]\)], "Input"],

Cell[BoxData[
    \(n = 6; set; mesh\ [ .001, 20, 1]\)], "Input"],

Cell[BoxData[
    \(n = 7; \ set; \ mesh[ .001, 30, 1]\)], "Input"],

Cell[BoxData[
    \(n = 8; \ set; \ mesh[ .001, 40, 1]\)], "Input"],

Cell[BoxData[
    \(n = 9; \ set; \ mesh[ .001, 40, 1]\)], "Input"],

Cell[BoxData[
    \(n = 10; \ set; \ mesh[ .001, 50, 1]\)], "Input"],

Cell[BoxData[
    \(n = 11; \ set; \ mesh[ .001, 50, 1]\)], "Input"],

Cell[BoxData[
    \(n = 12; \ set; \ mesh[ .001, 100, 1]\)], "Input"],

Cell[BoxData[
    \(n = 13; \ set; \ mesh[ .001, 100, 1]\)], "Input"],

Cell[BoxData[
    \(n = 14; \ set; \ mesh[ .001, 100, 1]\)], "Input"],

Cell[BoxData[
    \(n = 15; \ set; \ mesh[ .001, 100, 1]\)], "Input"],

Cell[BoxData[
    \(n = 16; \ set; \ mesh[ .001, 100, 1]\)], "Input"],

Cell[BoxData[
    \(End\)], "Input"]
}, Closed]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 1024}, {0, 768}},
AutoGeneratedPackage->Automatic,
CellGrouping->Manual,
WindowSize->{768, 490},
WindowMargins->{{102, Automatic}, {Automatic, 85}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 161, 5, 192, "Title"],
Cell[1903, 58, 106, 1, 69, "Subsubtitle"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2046, 64, 124, 3, 59, "Subsubsection"],
Cell[2173, 69, 475, 10, 86, "Text"],
Cell[2651, 81, 99, 2, 27, "Input",
  InitializationCell->True],
Cell[2753, 85, 2744, 69, 320, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[5534, 159, 129, 3, 45, "Subsubsection"],
Cell[5666, 164, 4830, 96, 683, "Input",
  InitializationCell->True],
Cell[10499, 262, 1941, 35, 427, "Input",
  InitializationCell->True],
Cell[12443, 299, 954, 16, 155, "Input"],
Cell[13400, 317, 4996, 92, 1035, "Input",
  InitializationCell->True],
Cell[18399, 411, 5857, 110, 1371, "Input",
  InitializationCell->True],
Cell[24259, 523, 71, 2, 43, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[24367, 530, 120, 3, 45, "Subsection"],
Cell[24490, 535, 3074, 55, 640, "Input",
  InitializationCell->True],
Cell[27567, 592, 187, 3, 30, "Text"],
Cell[27757, 597, 78, 1, 27, "Input"],
Cell[27838, 600, 79, 1, 27, "Input"],
Cell[27920, 603, 78, 1, 27, "Input"],
Cell[28001, 606, 78, 1, 27, "Input"],
Cell[28082, 609, 78, 1, 27, "Input"],
Cell[28163, 612, 82, 1, 27, "Input"],
Cell[28248, 615, 36, 1, 27, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[28321, 621, 185, 5, 45, "Subsection"],
Cell[28509, 628, 4396, 76, 930, "Input",
  InitializationCell->True],
Cell[32908, 706, 363, 8, 48, "Text"],
Cell[33274, 716, 65, 1, 27, "Input"],
Cell[33342, 719, 65, 1, 27, "Input"],
Cell[33410, 722, 67, 1, 27, "Input"],
Cell[33480, 725, 67, 1, 27, "Input"],
Cell[33550, 728, 67, 1, 27, "Input"],
Cell[33620, 731, 68, 1, 27, "Input"],
Cell[33691, 734, 68, 1, 27, "Input"],
Cell[33762, 737, 69, 1, 27, "Input"],
Cell[33834, 740, 69, 1, 27, "Input"],
Cell[33906, 743, 69, 1, 27, "Input"],
Cell[33978, 746, 69, 1, 27, "Input"],
Cell[34050, 749, 69, 1, 27, "Input"],
Cell[34122, 752, 36, 1, 27, "Input"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

